; Copyright (c) 2004-2010, Stubbe F. Hviid
;   All rights reserved
;+
; NAME:
;   FUNCTION osiris_subtract_bias, header, image, quality_map, sigma_map, prepix, postpix
;
; PURPOSE:
;   Remove coherent noise from image
;
; CATEGORY:
;   Osiris CCD Calibration
;
; CALLING SEQUENCE:
;  Result = osiris_calibrate_noise(header, image, quality_map, sigma_map, prepix, postpix)
;
; INPUTS:
;   header:   		The PDS image header of the image to be corrected
;   image:     		The image data of the image to be corrected
;   quality_map:	Existing map of quality estimate for each pixel to be modified
;   sigma_map:		Existing map of estimaped relative error for each pixel (in %) to be modified
;
; OPTIONAL INPUT:
;   prepix:     A prepixel stripe extracted from the image to be calibrated
;   postpix:	A postpixel stripe extracted from the image to be calibrated
;
; KEYWORD PARAMETERS:
;   PARAMETER_FILELIST: all external datafile names neede for this correction step
;                  will be added to this keywor parameter in a string array
;
; OUTPUTS:
;   header:   		The PDS image header of the image after correction
;   image:      	The image data of the image after correction
;   quality_map:	output map of quality estimate for each pixel
;   sigma_map:		output map of estimaped relative error for each pixel (in %)
;
; RETURNS:
;   0 on failure   1 on success
;
; RESTRICTIONS:
;   function requires that the paths defined in get_osiris_environment.pro has been set to correctly
;   point to the root directory of the external calibration files
;
;   The function depends on an external data file <camera>_<unit>_BIAS_V<version>_<timestamp>.dat
;   to exist in the BIAS folder of the database tree. This file must have a tag_file compatible structure
;   The file must contain tags with names of the form:
;                                   <frame mode>_<amblifier>_<binning>
;   As a minimum the file must contain the tag "default" giving the default bias level to use if no better
;   value can be found
;
; EXAMPLE:
;   p_read, pickfile(), h, d             ; read in raw PDS image
;   osiris_calibrate_noise, h, d
;
; MODIFICATION HISTORY:
;   Written by:  Stubbe F. Hviid   12/8-2004 based on p_adc_hi.pro by R. Kramm dated 29-OCT-2003
;                Sfh & MK          22/9-2005 Removed gt 0 statement from "(image - bias) > 0" allowing negative DN values
;                Sfh 02102006 - Switched to using osiris_model for detecting the camera and hardware model
;                MK 04/9-2008 - Corrections for amplifier both
;-

FUNCTION p_extract_start_time, filename
	fid1 = strmid(filename, 0, 1)
    fid3 = strmid(filename, 0, 3)
    len = strlen(filename)
    if (fid1 eq 'N' OR fid1 eq 'W') AND (len eq 52 OR len eq 30) then begin
            if fid3 eq 'NAC' OR fid3 eq 'WAC' then begin
            	; OSIRIS filenameing convention
       			return, p_scet2et(strmid(filename, 4, 23));
       		endif else begin
       			; PDS archive filenaming convention  21

				year = strmid(filename, 1, 4)
				month = strmid(filename, 5, 2)
				day = strmid(filename, 7, 2)
				hour = strmid(filename, 10, 2)
				minute = strmid(filename, 12, 2)
				sec = strmid(filename, 14, 2)
				msec = strmid(filename, 16, 3)

				return, julday(fix(month), fix(day), fix(year), fix(hour), fix(minute), fix(sec))
			endelse
	endif

    printf, -2, 'WARNING: Not a std filename - Cannot extract the SCET time from the filename - trying to read the header'
    p_rhead, filename, h
    return, p_scet2et(p_value(h, 'START_TIME'))
END


; -----------------------------------------------------------------------------
; --- Search for bias image in current directory
; ---
FUNCTION search_for_bias_image, header, bias_file
    htags = strupcase(TAG_NAMES(header))
    s = where(htags eq 'FILENAME')
    if s[0] eq -1 then begin
       printf, -2, 'WARNING: No filename specified in the header structure'
       return, -1
    endif

    file = header.filename

    if file eq '' then return, -1  ; No source directory defined - Do not know where to look

    fd = p_parse_file_path(file)

    camera = strmid(fd.name, 0, 1)

    if camera eq 'N' then begin
       mask = fd.dir + '/' + camera + '*F54.img'
    endif else begin
       mask = fd.dir + '/' + camera + '*F87.img'
    endelse

    files = findfile(mask)

    if files[0] eq '' then return, -1


    times = dblarr(n_elements(files))

	h_et = p_scet2et(p_value(header, 'START_TIME'))
    for i=0,n_elements(files)-1 do begin
       tmp =p_parse_file_path(files[i])
       times[i] = abs(p_extract_start_time(tmp.filename) - h_et)
    endfor

    s = sort(times)

    ; look for bias images closest in time to imag eto be calibrated
    for i=0, n_elements(files)-1 do begin
       p_read, files[s[i]], hb, db, /QUIET

       ; check for compatibility with current image
       if p_value(hb, 'SR_ACQUIRE_OPTIONS.EXPOSURE_DURATION') ne 0 then continue
       if p_value(hb, 'SR_ACQUIRE_OPTIONS.HARDWARE_BINNING_ID') ne p_value(header, 'SR_ACQUIRE_OPTIONS.HARDWARE_BINNING_ID') then continue
       if p_value(hb, 'SR_ACQUIRE_OPTIONS.WINDOWING_ENABLED_FLAG') ne p_value(header, 'SR_ACQUIRE_OPTIONS.WINDOWING_ENABLED_FLAG') then continue
       if p_value(hb, 'SR_ACQUIRE_OPTIONS.CRB_TO_PCM_SYNC_MODE') ne p_value(header, 'SR_ACQUIRE_OPTIONS.CRB_TO_PCM_SYNC_MODE') then continue
       if p_value(hb, 'SR_ACQUIRE_OPTIONS.AMPLIFIER_ID') ne p_value(header, 'SR_ACQUIRE_OPTIONS.AMPLIFIER_ID') then continue
       if p_value(hb, 'SR_ACQUIRE_OPTIONS.GAIN_ID') ne p_value(header, 'SR_ACQUIRE_OPTIONS.GAIN_ID') then continue
       if p_value(hb, 'SR_ACQUIRE_OPTIONS.ADC_ID') ne p_value(header, 'SR_ACQUIRE_OPTIONS.ADC_ID') then continue

       ; if this point is reached then every thing matches
       bs = where(db ne 0)

       ; set the bias file output variable
       tmp = p_parse_file_path(files[s[i]])
       bias_file = tmp.filename

       ; return measured bias
       return, median(float(db[bs]))
    endfor

    return, -1
END

FUNCTION p_bias_database_lookup, header, ampli, database_file
    ;ampli is the amplifier submitted from subtract_bias ('A' or 'B')
	osiris_model, header, camera, model
    ; amp may also be both
	amp    = p_value(header, 'SR_ACQUIRE_OPTIONS.AMPLIFIER_ID')
    bin    = p_value(header, 'SR_ACQUIRE_OPTIONS.HARDWARE_BINNING_ID')
    sync   = p_value(header, 'SR_ACQUIRE_OPTIONS.CRB_TO_PCM_SYNC_MODE')
    if p_value(header, 'SR_ACQUIRE_OPTIONS.WINDOWING_ENABLED_FLAG') eq 'TRUE' then begin
       fmode = 'WINDOW'
    endif else begin
       fmode = 'FULL'
    endelse

	; load required flatfield from database
	param_file = p_resolve_database_file('BIAS', camera, model, 'BIAS', '', 'TXT')
	if param_file eq '' then begin
		printf, -2, 'ERROR: Cannot resolve database file'
		return, -1
	endif

	; save the database filename to output
	database_file = param_file

	; read the database file
	params = fw_load_tagfile(param_file)

	; generate the tag
	if amp eq 'BOTH' then tag = strupcase(fmode + '_' + bin + '_' + amp + '_' + ampli + '_SYNC' + strtrim(sync, 2)) else $
	   tag = strupcase(fmode + '_' + bin + '_' + amp + '_SYNC' + strtrim(sync, 2))

	names = strupcase(TAG_NAMES(params))
	s = where(names  eq tag)

	if s[0] eq -1 then begin
		if file_test('osiris_cal_missing_bias.txt') eq 0 then out = [''] else out = p_read_text_file('osiris_cal_missing_bias.txt')
		out = [out, tag]
		p_write_text_file, 'osiris_cal_missing_bias.txt', out

		printf, -2, 'ERROR: Cannot find bias level for: ' + tag
		return, -1
	endif

	;print, 'Found database bias level for ' + camera + '.' + tag
	return, params.(s[0])
END


FUNCTION p_lookup_bias_level, header, amp, DATAFILE=DATAFILE

    ; first see if the image was acquired with prepixels downlinked
    ;fs = p_parse_file_path(header.filename)

    ;if file_test(fs.dir, /DIRECTORY) then begin
    ;   p = strpos(fs.filename, 'ID')
    ;   if p gt 0 then begin
    ;     DATAFILE = strmid(fs.filename, 0, p) + 'P' + amp + strmid(fs.filename, p+2)
    ;     file = fs.dir + '/' + DATAFILE
    ;     if file_test(file) then begin
    ;      p_read, file, hp, dp;

	;	  s = size(dp, /dim)
;		  xfrom = 0
;		  xto = s[0]-1
;		  if s[0] ge 5 then begin
;			xfrom += 2
;			xto   -= 2
;		  endif

;          return, mean(float(dp[xfrom:xto, 2:s[1]-1]))
;         endif
;       endif
;    endif

    ; if prepixels not found then look for bias image within the current directory
    ; with the same configuration as the image
;    level = search_for_bias_image(header, DATAFILE)
;    if level ge 0 then return, level

    ; Finally as the last option lookup the bias level from the database
    level = p_bias_database_lookup(header, amp, DATAFILE)
    return, level

END


FUNCTION osiris_subtract_bias, header, image, quality_map, sigma_map, prepix, postpix, PARAMETER_FILELIST=PARAMETER_FILELIST
VERSION='V2.1'

    if keyword_set(VERBOSE) then print, 'osiris_subtract_bias running'

    ; Verify the the correction has not already been performed
    flag = p_value(header, 'LEVEL2_PIPELINE.BIAS_CORRECTION_FLAG', /HISTORY)

    if flag eq 'TRUE' then begin
       if keyword_set(VERBOSE) then print, 'osiris_subtract_bias skipped - Already performed'
       return, 1
    endif

    ; do the calibration
    bias_level_msg = 'Undefined Bias Level'     ; initiate the bial level to undefined (-1)

    ; Extract some header information
    osiris_model, header, camera, model

    amp    = p_value(header, 'SR_ACQUIRE_OPTIONS.AMPLIFIER_ID')
    bin    = p_value(header, 'SR_ACQUIRE_OPTIONS.HARDWARE_BINNING_ID')
    sync   = p_value(header, 'SR_ACQUIRE_OPTIONS.CRB_TO_PCM_SYNC_MODE')
    if p_value(header, 'SR_ACQUIRE_OPTIONS.WINDOWING_ENABLED_FLAG') eq 'TRUE' then begin
       fmode = 'window'
    endif else begin
       fmode = 'full'
    endelse

    ; determine the relevant bias levels
    if amp eq 'BOTH' then begin
        if prepix[0] ge 0 then bias_a = median(prepix) else bias_a = p_lookup_bias_level(header, 'A', DATAFILE=datafile_a)
        if postpix[0] ge 0 then bias_b = median(postpix) else bias_b = p_lookup_bias_level(header, 'B', DATAFILE=datafile_b)

       if bias_a lt 0 OR bias_b lt 0 then begin
         printf, -2, 'ERROR: Cannot determine the bias level'
         return, 0
       endif

       ; subtract the bias level from the image data.
       ; Correction by MK. To be validated that this is indeed correct fo images that are not centered on pixel 1024 in x.

       dim = size(image, /DIM)
       xstart = p_value(header, 'IMAGE.FIRST_LINE_SAMPLE')-1
       if xstart+dim[0] le 1024 then image = image - bias_a $
       else if xstart ge 1024 then image = image - bias_a $
       else  begin
       image[0:1024-xstart-1, *] = (image[0:1024-xstart-1, *] - bias_a)
       image[1024-xstart:*, *]   = (image[1024-xstart:*,*]   - bias_b)
       end

       ; and write the bias level to the header as a vector
       p_setvalue, header, 'LEVEL2_PIPELINE.BIAS_LEVEL', [bias_a, bias_b], /HISTORY

       ; and add the external data file reference to the dependecy list
       if n_elements(datafile_a) ne 0 then begin
           if n_elements(PARAMETER_FILELIST) eq 0 then begin
               PARAMETER_FILELIST = [datafile_a]
           endif else begin
               PARAMETER_FILELIST = [PARAMETER_FILELIST, datafile_a]
    	   endelse
       endif
       if n_elements(datafile_b) ne 0 then begin
           if n_elements(PARAMETER_FILELIST) eq 0 then begin
               PARAMETER_FILELIST = [datafile_b]
           endif else begin
               PARAMETER_FILELIST = [PARAMETER_FILELIST, datafile_b]
    	   endelse
       endif
    endif else if amp eq 'A' then begin
       if prepix[0] ge 0 then bias_a = median(prepix) else bias_a = p_lookup_bias_level(header, 'A', DATAFILE=datafile)

       if bias_a lt 0 then begin
         printf, -2, 'ERROR: Cannot determine the bias level'
         return, 0
       endif

       ; subtract the bias level from the image data
       image = image - bias_a

       ; and write the bias level to the header as a scalar
       p_setvalue, header, 'LEVEL2_PIPELINE.BIAS_LEVEL', bias_a, /HISTORY

       ; and add the external data file reference to the dependecy list
       if n_elements(datafile) ne 0 then begin
           if n_elements(PARAMETER_FILELIST) eq 0 then begin
               PARAMETER_FILELIST = [datafile]
           endif else begin
               PARAMETER_FILELIST = [PARAMETER_FILELIST, datafile]
    	   endelse
       endif
    endif else begin
       if postpix[0] ge 0 then bias_b = median(postpix) else bias_b = p_lookup_bias_level(header, 'B', DATAFILE=datafile)

       if bias_b lt 0 then begin
         printf, -2, 'ERROR: Cannot determine the bias level'
         return, 0
       endif

       ; subtract the bias level from the image data
       image = image - bias_b

       ; and write the bias level to the header as a scalar
       p_setvalue, header, 'LEVEL2_PIPELINE.BIAS_LEVEL', bias_b, /HISTORY

       ; and add the external data file reference to the dependecy list
       if n_elements(datafile) ne 0 then begin
           if n_elements(PARAMETER_FILELIST) eq 0 then begin
               PARAMETER_FILELIST = [datafile]
           endif else begin
               PARAMETER_FILELIST = [PARAMETER_FILELIST, datafile]
    	   endelse
       endif
    endelse

    ; end calibration



    ; Change the processing flag
    p_setvalue, header, 'LEVEL2_PIPELINE.BIAS_CORRECTION_FLAG', 'TRUE', /LABEL, /NO_QUOTES, /HISTORY
    p_setvalue, header, 'LEVEL2_PIPELINE.BIAS_CORRECTION_VERSION', VERSION, /GROUP, /HISTORY


    ; finish
    if keyword_set(VERBOSE) then print, 'osiris_subtract_bias - Done'

    return, 1




END
