; Copyright (c) 2005, Stubbe F. Hviid
;   All rights reserved
;+
; NAME:
;   fw_load_tagfile
;
; PURPOSE:
;   Load a tag=value style text file into an IDL structure where the value elements are written in
;   valid IDL syntax. Comments are supported using the characters ';', '#' or '//' as the first character
;   in the line. The comments are discarded by the loader and will not be available in the output
;   structure.
;
; CATEGORY:
;   file io
;
; CALLING SEQUENCE:
;   FUNCTION fw_load_tagfile(file)
;
;
; INPUTS:
;   file: filename of file to be read
;
; OPTIONAL INPUTS:
;   None
;
; KEYWORD PARAMETERS:
;   QUIET:   If set no progress messges will be printed only error messages
;
; OUTPUTS:
;   An IDL structure where each tag in the tag=value list is turned into a structure tag of name <tag>
;
; OPTIONAL OUTPUTS:
;   None
;
; COMMON BLOCKS:
;   None
;
; SIDE EFFECTS:
;   None
;
; RESTRICTIONS:
;   None
;
; EXAMPLE:
;   Assume that the input file has the structure:
;     Hello="World"
;     Tag1=10
;   Then fw_load_tagfile() will return the IDL structure {Hello: "World", Tag1: 10}
;
; MODIFICATION HISTORY:
;   Written by:  Stubbe F. Hviid  June 2005

;-

FUNCTION fw_load_tagfile, file, QUIET=QUIET

    ; Tell the world that we are reading <file>
    ;if NOT keyword_set(QUIET) then print, 'Loading tagfile: ' + file

	; read the text file
	lines = p_read_text_file(file, /STRIP_PDS_HEADER, LABEL=LABEL)

	; strip comment lines
	for i=0, n_elements(lines)-1 do begin
		if strmid(strtrim(lines[i], 1), 0, 1) eq '#' then lines[i] = ''
	endfor
	s = where(lines ne '')
	if s[0] ne -1 then lines = lines[s] else lines = ['']


    ;if NOT keyword_set(QUIET) then print, 'Found ' + strtrim(n_elements(lines)-1, 2) + ' tags in file'

    ; Check for empty input
    if n_elements(lines) eq 0 then begin
        print, 'Empty input from file: ' + file
        return, -1
    endif

    ; Then build structure source line
    out = {tagfile: file, label: LABEL}
    for i=0, n_elements(lines)-1 do begin
       p = strpos(lines[i], '=')
       if p gt 0 then begin
       	 tag = strmid(lines[i], 0, p)
       	 value = strmid(lines[i], p+1)

		 value = byte(value)
		 s = where(value eq 34)
		 if s[0] ne -1 then value[s] = 39
		 value = string(value)


		 if execute('rvalue = ' + value) eq 0 then begin
			stop
			return, -1
		 endif

		 out = p_set_struct_tag(out, tag, rvalue)
	   endif

    endfor

	return, out

    ; Error handling
    ioe:
    print, 'Error loading tag file: ' + file
    return, -1


END